## Overview
This repository contains python scripts to re-create the result shown in

`Error bounds for kernel-based approximations of the Koopman operator, arxiv:2301.08637`

We consider a one-dimensional Ornstein-Uhlenbeck process

$ dX_t = -\alpha X_t + \sqrt{2\beta^{-1}} dW_t $

where $\alpha$ is a spring constant and $\beta$ is inverse temperature.
As approximating RKHS, we use the one generated by the Gaussian RBF kernel.
We compare error estimates and empirical results for three bandwidths $ \sigma \in \{0.05, 0.1, 0.5 \}$, and for a range of data sizes $m$ between $m = 20$ and $m = 50.000$. 

All results are saved to the subdirectory `Data/`, figures are saved to `Figures/`. To re-create the results from scratch, run `analytical_quantities.py`, `theoretical_bounds.py`, and `data_errors.py` in that order. Generate the figures by running
`figures.py`.


#### `analytical_quantities.py`
In this script,, we calculate three fundamental quantities:
1.  matrix of Koopman matrix elements for Mercer features:
$$ M_{ij} = \langle \phi_i, \, \mathcal{K}^t \phi_j\rangle_\mu, $$

2. the Hilbert-Schmidt norm of the cross-covariance operator, using the formula
$$ \|C^t_\mathbb{H} \|^2_{HS} = \sum_{i, j} \lambda_i \lambda_j M_{ij}^2, $$
where $(\lambda_i, \phi_i)$ are eigenpairs of the kernel integral operator.
The series is truncated after $n_{\max} = 10$ terms in both sums.

3. The quantities $d_{r, t}$ for all Koopman eigenfunctions $\psi_r$, according to the formula:
$$ d_{r, t} = \sum_{i,j} \lambda_i, \lambda_j \langle \phi_i \psi_r,\, \mathcal{K}^t \phi_j \rangle^2_\mu.$$
The series is truncated after $n_{\max} = 20$ terms for $\sigma=0.05$ and $n_{\max} = 10$ terms otherwise. We compute $d_{r, t}$ for $r \leq J_{\max} = 15$. The results are stored to file as the calculation takes a few minutes for each bandwidth. The resulting files are called `AnalyticalQuantities_t=*_sigma=*.npz`.


#### `theoretical_bounds.py`
In this script, we calculate theoretical error bounds for all data sizes $m$ and all bandwidths $\sigma$.

First, we calculate several bounds for the estimation error of the cross-covariance operator, based on the Chebyshev-inequality:
$$ E(m) = \left[\frac{\sigma^2}{\delta m}. \right]^{1/2}$$
The variance $\sigma^2$ is calculated in different ways

1. Using the i.i.d. variance:
$$ \sigma^2 = \mathbb{E}_0(t) = 1 - \|C^t_\mathbb{H} \|^2_{HS}. $$
2. Using the ergodic variance (Theorem 3.1) and eigenvalues $q_r$ of the Koopman operator:
$$ \sigma^2 = \mathbb{E}_0(t) + \sum_{r=1}^\infty d_{r, t} F_m(q_r), \\
F_m(q) = \frac{2}{1 - q} \,\frac{1 - (1 - q^m)}{m (1 - q)}, $$
3. the bound from Remark 3.4:
$$ \sigma^2 = \mathbb{E}_0(t) \frac{3 - q_1}{1 - q_1} $$

Second, we also compute the resulting prediction error for the empirical and rank-truncated Koopman operator (at rank $n_{\max} = 10$), according to Theorem 4.1 and 4.5. Using the error $E(m)$ computed in the first step, the error estimate is:
$$ \left[\frac {1}{\sqrt{\lambda_{n_{\max}}}} + \frac{2(n_{\max}+1)}{\delta_N\lambda_{n_{\max}}}\right]E(m). $$
The results are saved to file as `Error_Bounds_t=*_sigma=*.npz`

#### `data_errors.py`
In this part, we empirically determine the estimation error for the cross-covariance operator,
and the prediction error for a Gaussian observable using a finite number of simulations.

For the estimation error, we generate 200 independent simulations of the total length equal to
the maximal data size $m$ times the lag time $t$. For each simulation and all data sizes
considered above, we compute the estimation error based on the formula:

$$ E(m)^2 = \|C^t_\mathbb{H} - \hat{C}^t_\mathbb{H} \|^2_{HS} = \|C^t_\mathbb{H}\|^2 - 2\langle C^t_\mathbb{H},\, \hat{C}^t_\mathbb{H} \rangle_{HS} + \|\hat{C}^t_\mathbb{H} \|^2_{HS}$$

To calculate the individual contributions, we use the formulae (see above):

$$\begin{aligned}
\|C^t_\mathbb{H} \|^2_{HS} &= \sum_{i, j} \lambda_i \lambda_j M_{ij}^2, & \|\hat{C}^t_\mathbb{H} \|^2_{HS} &= \sum_{i, j} \lambda_i \lambda_j \hat{M}_{ij}^2, & \langle C^t_\mathbb{H},\, \hat{C}^t_\mathbb{H} \rangle_{HS} &= \sum_{i, j} \lambda_i \lambda_j M_{ij}\hat{M}_{ij},
\end{aligned}$$

where

$$\begin{aligned}
M_{ij} &= \langle \phi_i, \mathcal{K}^t \phi_j \rangle_\mu, & \hat{M}_{ij} &= \frac{1}{m}\sum_{k=1}^{m} \phi_i(x_k) \phi_j(x_{k+1}).
\end{aligned}$$

The matrix $M$ has been pre-computed above, while $\hat{M}$ is calculated from the data on the fly.

To compute the prediction error, we make use of the fact that for a Gaussian observable $f$ with variance 
$\sigma_0^2$ and mean value $\mu_0$, the action of the Koopman operator leads to the analytical result

$$ \mathcal{K}^t f(x) = \frac{1}{\sqrt{2\pi\sigma_t^2}} \exp\left(-\frac{(\mu_0 - e^{-\alpha t}x)^2}{2\sigma_t^2}\right), $$

where $\sigma_t ^2 = \sigma_0^2 + \frac{1 - \exp(-2\alpha t)}{\alpha\beta}$. We calculate the truncated Mercer
series expansion of this function by computing its $L^2_\mu$-inner products with the Mercer features. This 
truncated expansion serves as the reference for prediction.

We then generate 20 independent simulations, and diagonalize the kernel Gramian matrix to obtain
empirical Mercer features. The application of the empirical (truncated) Koopman operator to the observable
$f$ is performed as evaluated by the formula 

$$\hat{K}^{m, t}_N f(x) 
= \frac{1}{m} f(Y)^\top V_N \Lambda_N^{-1} V_N^\top K_{\mathcal{X}, x}. $$

Here, $Y$ is the time-lagged data, $V_N$ and $\Lambda_N$ are the matrices of eigenvectors and eigenvalues of the empirical kernel integral operator, and $K_{\mathcal{X}, x}$ is the vector of kernel evaluations at the data sites $\mathcal{X}$ and the position $x$.

This empirical observable is compared to the reference
by computing their $L^2_\mu$-error using numerical integration. The results are stored as `Empirical_Errors_t=*_sigma=*.npz`.

#### `figures.py`
This function gathers the results of the previous three steps and produces the plots shown in the paper.

#### `OU.py`
System class to represent Ornstein-Uhlenbeck process. Provides methods to generate simulation data, evaluate
the potential, force, and invariant density of the system.

#### `gaussian_kernel.py`
Class to represent Gaussian kernel. Can be used to evaluate the gaussian kernel on pairs of data matrices.

#### `mercer.py`
Provides a class that stores all constants and eigenpairs of the Mercer features of the Gaussian kernel for the OU process.