import numpy as np
from scipy.special import gamma, eval_hermite

class Mercer:
    def __init__(self, sigma, alpha):
        self.sigma = sigma
        self.alpha = alpha

        # Constants related to Mercer features:
        self.eta = (1 + 4.0/(self.sigma**2 * self.alpha))**(1/4)
        self.gamma_n = lambda ll: (self.eta / (2**ll * gamma(ll + 1)))**(1/2)
        self.zeta = np.sqrt(0.5 * self.alpha * (self.eta**2 - 1.0))
        # Eigenvalues of the kernel integral operator:
        self.C1 = self.alpha + self.zeta**2 + (1.0/self.sigma)**2
        self.lambda_n = lambda ll: np.sqrt(self.alpha/self.C1) * (1.0 / (self.sigma**2 * self.C1))**ll
        # ... associated eigenfunctions:
        self.phi = lambda x, n: (self.gamma_n(n) * np.exp(-self.zeta**2 * x**2)
                                  * eval_hermite(n, np.sqrt(self.alpha)*self.eta * x))